#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-07 21:04:20 (ywatanabe)"

from ._torch_fn import torch_fn

__copyright__ = "Copyright (C) 2024 Yusuke Watanabe"
__version__ = "0.0.2"
__license__ = "MIT"
__author__ = "ywatanabe1989"
__author_email__ = "ywata1989@gmail.com"
__url__ = "https://github.com/ywatanabe1989/torch_fn"
